/* PROGRAM TABOUT 2.7 converts tabs to spaces
usage : TABOUT.exe Myfile.TXT
        - expands tabs to spaces
        - trims off ^z Eof char
        - ensures each line ends in CrLf
        - trims trailing spaces of each line.
*/
/*
Roedy Green
Canadian Mind Products
#101 - 2536 Wark Street
Victoria, BC Canada V8T 4G8
tel:(250) 361-9093
mailto:roedyg@mindprod.com
http://mindprod.com
*/
/*
Windows console application.
For Visual C++ Express options see http://mindprod.com/jgloss/cpp.html#COMPILING
Version History:
 2.0 1990-04-29
 2.1 1993-06-08 embed new address and phone
 2.2 1996-10-25 embed Quathiaski address.
                simplify code to create temporary workfile.
                simplify code to check extensions.
                new algorithm that does not have a maximum line length
 2.3 1997-03-06 add *.SQL, *.HTM, *.INI, *.DLL file types
 2.4 1998-11-08 embed Barker address
 2.5 2008-02-03 convert from DOS to Windows to handle long file names
                use safe character library
                POSIX compliant
                display banner on all errors
                add icon, pad, docs
 2.6 2009-03-12 add ANT build script
 2.7 2010-01-16 build in more good/bad extensions
*/
/*
The program will expand tabs from files with the following extensions:
        .ASM .PAS .BAT .CTL .CMD .LST .MAC .TXT .USE .KEY .ANS .C .H etc.
The program will refuse to expand tabs from files with the following
extensions:
    bad =.EXE .COM .OBJ etc.
For all other extensions the program will prompt with a warning
message ask and for confirmation before continuing.  *.DOC are
often MS-Word files - which must be left alone.
Source and excutables may be freely used for any purpose except military.
*/
/* ==================================== */
#define Esc '\x1B'
#define TabSpacing 8
#include <stdlib.h>  /* exit, putenv, _splitpath */
#include <stdio.h>   /* fclose, fgetc, printf, remove, rename, setvbuf */
#include <string.h>  /* strcat, strcmp, strupr */
#include <conio.h>   /* getch, putch */
#include <ctype.h>   /* isUpper isLower toUpper */
/* ==================================== */
/* use all global variables and no parameter passing for simplicity. */
static FILE *Before;
/* input file containing no ^Z chars, except possibly at the end */
static FILE *After;
/* output file with tabs expanded to tabs */
static char * BFilename;
/* name of file we will convert */
static char * AFilename;
/* name of the temporary output file */
/* ==================================== */
/* P R O T O T Y P E S  */
int  main (int argc,  char *argv[]);
void Banner (void);
void Die (void);
void Honk (void);
void OpenAFile (void);
void OpenBFile (void);
void SafeFilename (void);
/* ==================================== */
int main( int argc,  char *argv[] )
   /* main TABOUT */
   {
   int c; /* process char by char */
   int col = 0; /* 1-based column we have written */
   /* 0 = no non-blank chars on line yet */
   int pendingSpaces = 0;  /* spaces we may later drop or emit */
   int i; /* local loop counter */
   if ( argc != 2 /*  0=TabOut.Exe 1=MyFile.Txt  */ )
      {
      Banner();
      printf("Oops!  usage:  TABOUT  Myfile.TXT");
      printf("Only one file allowed on the command line.\n");
      Die();
      }
   BFilename = *++argv; /*  want first arg, not progname  */
   OpenBFile();    /* Open input "before" file. */
   /* Make sure file exists before */
   /* song and dance about extension. */
   SafeFilename(); /* make sure filename has sane extension */
   OpenAFile();    /* open output "after" file */
   printf("Converting tabs to spaces in %s ...\n",BFilename);
   while ( (c=fgetc(Before)) != EOF )
      {
      switch ( c )
         {
         case ' ':
            /* save up spaces, so trailing spaces can be dropped. */
            ++pendingSpaces;
            break;
         case '\t':
            /* effectively expand tab to 1 to 8 spaces */
            pendingSpaces += (TabSpacing - ((col+pendingSpaces) % TabSpacing));
            break;
         case '\n':
            /* ignore pending/trailing spaces */
            fputc('\n',After);
            /* runtime library expands to CrLf */
            col = 0;
            pendingSpaces = 0;
            break;
         default:
            /* ordinary non-blank char */
            /* squirt out any pent up spaces */
            for ( i=0; i<pendingSpaces; i++ )
               {
               fputc(' ',After);
               }
            col += pendingSpaces;
            pendingSpaces = 0;
            /* put out the char itself */
            fputc(c,After);
            col++;
            break;
         } /* end switch */
      } /* end while */
   /* Rename output to input */
   fclose (Before);
   fclose (After);
   remove (BFilename);
   rename (AFilename, BFilename);
   return(0);
   } /* main TABOUT */
/* ==================================== */
void Banner(void)
   {
   /* display copyright banner, with shaded splashes */
   printf("\n" "\xb0\xb1\xb2\xdb" " Tabout 2.7 "  "\xdb\xb2\xb1\xb0"
          "\nFreeware to convert tabs to spaces."
          "\nCopyright: (c) 1990-2017 Roedy Green, Canadian Mind Products"
          "\n" "#101 - 2536 Wark Street, Victoria, BC Canada V8T 4G8"
          "\n" "tel:(250) 361-9093   mailto:roedyg@mindprod.com   http://mindprod.com"
          "\n" "May be used freely for non-military use only"
          "\n\n");
   } /* Banner */
/* ==================================== */
void SafeFilename(void)
   {
   /* Ensure appropriate file name extensions.
      good =.ASM .PAS .etc - done without prompt
       bad =.EXE .COM .OBJ - abort
   warning =.DOC & others
   */
   static const char  * GoodExtensions  [] =
   {
       ".ANS",
       ".ASM",
       ".BAT",
       ".C",
       ".CMD",
       ".CPP",
       ".CSV",
       ".CTL",
       ".H",
       ".HPP",
       ".HTM",
       ".HTML",
       ".IH",
       ".INI",
       ".JAVA",
       ".JNLP",
       ".KEY",
       ".LST",
       ".MAC",
       ".PAS",
       ".RH",
       ".SQL",
       ".TXT",
       ".USE",
       ".XML",
        0};
   /* 0 is just end marker */
   static const char * BadExtensions [] =
   {
       ".CLASS",
       ".COM",
       ".DLL",
       ".EXE",
       ".GIF",
       ".JAR",
       ".JPG",
       ".MSI",
       ".OBJ",
       ".PDF",
       ".PNG",
       ".SER",
       ".ZIP",
       0};
   int Response; /* Y or N, yes Virginia, int, C is weird */
   char Extension[_MAX_EXT];
   int i; /* local loop counter */
   _splitpath_s( BFilename, NULL /* drive */, 0, NULL /* dir */, 0, NULL /* name */, 0, Extension, _MAX_EXT);
   _strupr_s(Extension, _MAX_EXT); /* convert to upper case for compare */
   for ( i=0 ; GoodExtensions[i]; i++ )
      {
      if ( strcmp(Extension,GoodExtensions[i])==0 )
         { /* match, it is Good */
         return;
         }
      }
   for ( i=0 ; BadExtensions[i] ; i++ )
      {
      if ( strcmp(Extension,BadExtensions[i])==0 )
         { /* match, it is bad */
         Banner();
         printf("Oops!  Tabout cannot be used on EXE COM or OBJ files "
                "such as %s\n",
                BFilename);
         Die();
         }
      }
   /* just give a warning */
   printf("Warning!\n"     /* new line to give room for long filename */
          "Tabout is not usually used on %s files such as %s\n",
          Extension,BFilename);
   printf("Do you want to continue and expand the tabs anyway?"
          "  (Y)es (N)o  ");
   while ( 1 ) /* loop forever till user enters Y or N */
      {
      Honk();
      Response = _getch();
      /* not echoed because user might hit tab or Enter */
      /* and mess up the screen */
      Response = toupper(Response);
      /* toupper is a macro, so needs simple argument */
      switch ( Response )
         {
         case 'Y':
            printf("Yes\n");
            return;
         case 'N':
            printf("No\n");
            /* fallthru */
         case Esc :
            printf("\nTabout aborted\n");
            Die();
            /* others, keep looping */
         }
      }
   }  /* SafeFileName */
/* ==================================== */
void OpenBFile(void)
   { /* open input Before file */
   if ( fopen_s(&Before, BFilename, "rt") )
      {
      Banner();
      printf("Oops!  Cannot find file %s\n",BFilename);
      Die();
      }
   setvbuf(Before,NULL,_IOFBF,40*512);
   } /* OpenBFile */
/* ==================================== */
void OpenAFile(void)
   {
   /* create a uniquely named temporary file to hold converted text */
   /* It must be in the same directory as the BFilename since we will */
   /* rename it later. */
   char drive[_MAX_DRIVE]; /* with colon */
   char dir[_MAX_DIR];     /* with lead/trail \ */
   char name[_MAX_FNAME];
   char ext[_MAX_EXT];     /* with lead . */
   char filepath[_MAX_DRIVE + _MAX_DIR];
   _splitpath_s(BFilename, drive, _MAX_DRIVE, dir,_MAX_DIR, name, _MAX_FNAME, ext, _MAX_EXT);
   strcpy_s(filepath, _MAX_DRIVE + _MAX_DIR, drive);
   strcat_s(filepath, _MAX_DRIVE + _MAX_DIR , dir);
   /* Force to current directory if empty */
   if ( strcmp(filepath, "") == 0 )
      strcpy_s(filepath,_MAX_DRIVE + _MAX_DIR, ".");
   else
      filepath[strlen(filepath) - 1] = 0;  /* chop trail \ */
   /* stop tempnam from using SET TMP directory instead of filepath */
   /* Sets TMP just for this and any "children" */
   /* processes -- doesn't change parent's TMP */
   _putenv("TMP=");
   if ( (AFilename = _tempnam(filepath, "TMP")) == NULL )
      {
      Banner();
      printf("Oops!  Cannot create the temporary work file\n\a");
      exit(1);
      }
   if ( fopen_s(&After, AFilename, "wt") )
      {
      Banner();
      printf("Oops! Cannot open work file %s\n\a", AFilename);
      exit(1);
      }
   setvbuf(After,NULL,_IOFBF,40*512);
   } /* end OpenAFile */
/* ==================================== */
void Honk (void)
   {
   /* make a noise */
   printf("\a");
   }
/* =================================== */
void Die (void)
   {
   Honk();
   if ( Before )
      {
      fclose ( Before );
      }
   if ( After )
      {
      fclose ( After );
      }
   exit(1);   /* exit with errorlevel = 1 */
   } /* Die */
/* ==================================== */
/* -30- */
